from IPython.display import HTML
from markdown import Markdown

from mkapi.core.base import Node
from mkapi.core.node import get_node

converter = Markdown()


def get_html(node):
    if not isinstance(node, Node):
        node = get_node(node)
    markdown = node.get_markdown()
    html = converter.convert(markdown)
    node.set_html(html)
    return node.render()


def display(name):
    return HTML(get_html(name))



"""This module provides entity classes to represent docstring structure."""
from dataclasses import dataclass, field
from typing import Any, Iterator, List, Optional

import mkapi.core.preprocess
from mkapi.core.signature import Signature
from mkapi.core.renderer import renderer


@dataclass
class Base:
    """Base class.

    Args:
        name: Object name.
        type: Object type.
        markdown: Markdown source.

    Attributes:
        name: Object name.
        type: Object type.
        type_html: Object type html after adding links.
        markdown: Markdown source.
        html: HTML after conversion.
    """

    name: str
    type: str = ""
    markdown: str = ""
    html: str = field(default="", init=False)
    type_html: str = field(default="", init=False)

    def set_html(self, html: str):
        """Sets `html` attribute."""
        self.html = html


@dataclass
class Item(Base):
    """Item class represents an item in Parameters, Attributes, and Raises sections."""

    def set_html(self, html: str):
        """Sets `html` attribute cleaning `p` tags."""
        html = html.replace("<p>", "").replace("</p>", "<br>")
        if html.endswith("<br>"):
            html = html[:-4]
        self.html = html


@dataclass
class Section(Base):
    """Section class represents a section in docstring.

    Args:
        items: List for Arguments, Attributes, or Exceptions sections.

    Attributes:
        items: List for Arguments, Attributes, or Exceptions sections.

    Examples:
        `Section` is iterable:
        >>> section = Section('Returns', markdown='Returns an integer.')
        >>> for x in section:
        ...     assert x is section

        >>> items = [Item('x'), Item('y'), Item('z')]
        >>> section = Section('Parameters', items=items)
        >>> [item.name for item in section]
        ['x', 'y', 'z']

        Indexing:
        >>> isinstance(section['y'], Item)
        True
        >>> section['z'].name
        'z'

        Contains:
        >>> 'x' in section
        True
    """

    items: List[Item] = field(default_factory=list)

    def __iter__(self) -> Iterator[Base]:
        if self.markdown:
            yield self
        else:
            yield from self.items

    def __getitem__(self, name) -> Optional[Item]:
        for item in self.items:
            if item.name == name:
                return item
        return None

    def __contains__(self, name) -> bool:
        return self[name] is not None


@dataclass
class Docstring:
    """Docstring class represents a docstring of an object.

    Args:
        sections: List of Section instance.
        type: Type for Returns and Yields sections.

    Attributes:
        sections: List of Section instance.
        type: Type for Returns and Yields sections.

    Examples:
        Empty docstring:
        >>> docstring = Docstring()
        >>> assert not docstring

        Docstring with 3 sections:
        >>> default = Section("", markdown="Default")
        >>> parameters = Section("Parameters", items=[Item("a"), Item("b")])
        >>> returns = Section("Returns", markdown="Returned Value")
        >>> docstring = Docstring([default, parameters, returns])

        `Docstring` is iterable:
        >>> [base.name for base in docstring]
        ['', 'a', 'b', 'Returns']

        Indexing:
        >>> docstring["Parameters"].items[0].name
        'a'
    """

    sections: List[Section] = field(default_factory=list)
    type: str = ""

    def __bool__(self):
        return len(self.sections) > 0

    def __iter__(self) -> Iterator[Base]:
        for section in self.sections:
            yield from section

    def __getitem__(self, name: str) -> Optional[Section]:
        for section in self.sections:
            if section.name == name:
                return section
        return None

    def __setitem__(self, name: str, section: Section):
        for k, section_ in enumerate(self.sections):
            if section_.name == name:
                self.sections[k] = section
                return
            if name == "Parameters" and section_.name != "":
                self.sections.insert(k, section)
                return
            if section_.name in ["Raises", "Yields", "Returns"]:
                self.sections.insert(k, section)
                return
        self.sections.append(section)


@dataclass
class Node(Base):
    """Node class represents an object.

    Attributes:
        obj: Object.
        depth: Current depth of object searching.
        prefix: Prefix.
        kind: Kind such as `function`, `class`, `module`, etc.
        sourcefile: Souce filename thats defines this object.
        lineno: Line number.
        signature: Signature instance.
        docstring: Docstring instance.
        members: Member objects. For example, methods of class.
        headless: Used in page mode.
        html: HTML after rendering.
    """

    obj: Any = field(default=None, repr=False)
    depth: int = 0
    prefix: str = ""
    kind: str = ""
    sourcefile: str = ""
    lineno: int = 0
    signature: Signature = field(default_factory=lambda: Signature())
    docstring: Docstring = field(default_factory=lambda: Docstring())
    members: List["Node"] = field(default_factory=list)
    headless: bool = False
    prefix_url: str = ""
    name_url: str = ""

    def __post_init__(self):
        if self.prefix:
            self.id = ".".join([self.prefix, self.name])
        else:
            self.id = self.name

    def __getitem__(self, index):
        return self.members[index]

    def __len__(self):
        return len(self.members)

    def __getattr__(self, name):
        for member in self.members:
            if member.name == name:
                return member

    def __iter__(self) -> Iterator[Base]:
        if self.docstring:
            yield from self.docstring
        for member in self.members:
            yield from member

    def get_markdown(self) -> str:
        """Returns a Markdown source for docstring of this object."""
        markdowns = []
        for base in self:
            markdown = mkapi.core.preprocess.convert(base.markdown)
            markdowns.append(markdown)
        return "\n\n<!-- mkapi:sep -->\n\n".join(markdowns)

    def set_html(self, html: str):
        """Sets HTML to `Base` instances recursively."""
        for base, html in zip(self, html.split("<!-- mkapi:sep -->")):
            base.set_html(html.strip())

    def render(self) -> str:
        self.html = renderer.render(self)
        return self.html



import atexit
import importlib
import inspect
import logging
import os
import shutil
import sys
from typing import Iterator, List, Tuple

from mkapi.core.node import get_node
from mkapi.core.renderer import renderer

logger = logging.getLogger("mkdocs")


def create_nav(config):
    nav = config["nav"]
    docs_dir = config["docs_dir"]
    config_dir = os.path.dirname(config["config_file_path"])
    pages_all = []
    for page in nav:
        if isinstance(page, dict):
            for key, value in page.items():
                if isinstance(value, str) and value.startswith("mkapi/"):
                    page[key], pages = walk(value, docs_dir, config_dir)
                    pages_all.extend(pages)
    return config, pages_all


def _walk(top: str) -> Iterator[List[str]]:
    for root, dirs, files in os.walk(top):
        paths = [root]
        for x in dirs:
            if x.startswith("__"):
                dirs.remove(x)
        for file in files:
            if file.endswith(".py") and not file.startswith("_"):
                paths.append(file[:-3])
        yield paths


def walk(value: str, docs_dir: str, config_dir) -> Tuple[list, list]:
    _, api_path, *tops = value.split("/")
    abs_api_path = os.path.join(docs_dir, api_path)
    if os.path.exists(abs_api_path):
        logger.error(f"[MkApi] {abs_api_path} exists: Delete manually for safety.")
        sys.exit(1)
    os.mkdir(abs_api_path)
    atexit.register(lambda path=abs_api_path: rmtree(path))

    top = os.path.join(config_dir, *tops)
    root = os.path.dirname(top)

    if root not in sys.path:
        sys.path.insert(0, root)
    nav = []
    pages_all = []
    for paths in _walk(top):
        package = os.path.relpath(paths[0], root)
        package = package.replace("/", ".").replace("\\", ".")
        package_obj = importlib.import_module(package)
        if inspect.getdoc(package_obj):
            paths[0] = ""
        else:
            paths = paths[1:]

        pages = []
        for path in paths:
            if path:
                module = ".".join([package, path])
                module_obj = importlib.import_module(module)
                if not inspect.getdoc(module_obj):
                    continue
            else:
                module = package
            abs_path = os.path.normpath("/".join([abs_api_path, module]) + ".md")
            if path == "":
                children = paths[1:]
            else:
                children = []
            create_page(abs_path, module, children)
            page = os.path.relpath(abs_path, docs_dir).replace("\\", "/")
            pages_all.append(abs_path)
            pages.append({page[:-3].split(".")[-1]: page})
        if pages:
            nav.append({package: pages})

    return nav, pages_all


def rmtree(path):
    if not os.path.exists(path):
        return
    try:
        shutil.rmtree(path)
    except PermissionError:
        logger.warning(f"[MkApi] Couldn't delete directory: {path}")


def create_page(abs_path, module: str, children: List[str]):
    with open(abs_path, "w") as f:
        f.write(create_markdown(module, children))


def create_markdown(module: str, children: List[str]) -> str:
    node = get_node(module, max_depth=1)
    members = [member.id for member in node.members]
    markdown = renderer.render_page(node, module, members, children)
    return markdown



"""This module provides functions that parse docstring."""

import inspect
import re
from typing import Any, Iterator, List, Tuple

from mkapi.core.base import Docstring, Item, Section
from mkapi.core.signature import Signature

SECTIONS = [
    "Args",
    "Arguments",
    "Attributes",
    "Example",
    "Examples",
    "Note",
    "Notes",
    "Parameters",
    "Raises",
    "Returns",
    "References",
    "See Also",
    "Todo",
    "Warning",
    "Warnings",
    "Warns",
    "Yield",
    "Yields",
]


def get_indent(line: str) -> int:
    indent = 0
    for x in line:
        if x != " ":
            return indent
        indent += 1
    return -1


def join(lines):
    indent = get_indent(lines[0])
    return "\n".join(line[indent:] for line in lines).strip()


def rename_section(name: str) -> str:
    if name in ["Args", "Arguments"]:
        return "Parameters"
    if name == "Warns":
        return "Warnings"
    return name


def section_header(line: str) -> Tuple[str, str]:
    """Returns a tuple of (section name, style name).

    Args:
        line: Docstring line.

    Examples:
        >>> section_header("Args:")
        ('Args', 'google')
        >>> section_header("Raises")
        ('Raises', 'numpy')
        >>> section_header("other")
        ('', '')
    """
    if line in SECTIONS:
        return line, "numpy"
    elif line.endswith(":") and line[:-1] in SECTIONS:
        return line[:-1], "google"
    else:
        return "", ""


def split_section(doc: str) -> Iterator[Tuple[str, str, str]]:
    """Yields a tuple of (section name, contents, style).

    Args:
        doc: Docstring

    Examples:
        >>> doc = "abc\\n\\nArgs:\\n    x: X\\n"
        >>> it = split_section(doc)
        >>> next(it)
        ('', 'abc', '')
        >>> next(it)
        ('Parameters', 'x: X', 'google')
    """
    lines = [x.rstrip() for x in doc.split("\n")]
    name = ""
    style = ""
    start = indent = 0
    for stop, line in enumerate(lines, 1):
        if stop == len(lines):
            next_indent = -1
        else:
            next_indent = get_indent(lines[stop])
        if not line and next_indent < indent and name:
            if start < stop - 1:
                yield name, join(lines[start : stop - 1]), style
            start = stop
            name = ""
        else:
            section, style_ = section_header(line)
            if section:
                if start < stop - 1:
                    yield name, join(lines[start : stop - 1]), style
                style = style_
                name = rename_section(section)
                start = stop
                if style == "numpy":  # skip underline without counting the length.
                    start += 1
                indent = next_indent
    if start < len(lines):
        yield name, join(lines[start:]), style


def split_parameter(doc: str) -> Iterator[List[str]]:
    """Yields list of parameter string.

    Args:
        doc: Docstring
    """
    lines = [x.rstrip() for x in doc.split("\n")]
    start = stop = 0
    for stop, line in enumerate(lines, 1):
        if stop == len(lines):
            next_indent = 0
        else:
            next_indent = get_indent(lines[stop])
        if next_indent == 0:
            yield lines[start:stop]
            start = stop


def parse_parameter(lines: List[str], style: str) -> Tuple[str, str, str]:
    """Yields (name, type, markdown).

    Args:
        lines: Splitted parameter docstring lines.
        style: Docstring style. `google` or `numpy`.
    """
    if style == "google":
        name, _, line = lines[0].partition(":")
        name = name.strip()
        parsed = [line.strip()]
        pattern = r"(.*?)\s*?\((.*?)\)"
    else:
        name = lines[0].strip()
        parsed = []
        pattern = r"([^ ]*?)\s*:\s*(.*)"
    if len(lines) > 1:
        indent = get_indent(lines[1])
        for line in lines[1:]:
            parsed.append(line[indent:])
    m = re.match(pattern, name)
    if m:
        name, type = m.group(1), m.group(2)
    else:
        type = ""
    return name, type, "\n".join(parsed)


def parse_parameters(doc: str, style: str) -> List[Tuple[str, str, str]]:
    """Returns list of (name, type, markdown)."""
    return [parse_parameter(lines, style) for lines in split_parameter(doc)]


def parse_raise(lines: List[str], style: str) -> Tuple[str, str]:
    """Returns (type, markdown)."""
    type, _, markdown = parse_parameter(lines, style)
    return type, markdown


def parse_raises(doc: str, style: str) -> List[Tuple[str, str]]:
    """Returns list of (type, markdown)."""
    return [parse_raise(lines, style) for lines in split_parameter(doc)]


def parse_returns(doc: str, style: str) -> Tuple[str, str]:
    """Returns (type, markdown)."""
    lines = doc.split("\n")
    if style == "google":
        if ":" in lines[0]:
            type, _, lines[0] = lines[0].partition(":")
            type = type.strip()
            lines[0] = lines[0].strip()
        else:
            type = ""
    else:
        type = lines[0].strip()
        lines = lines[1:]
    return type, join(lines)


def create_section(name: str, doc: str, style: str) -> Section:
    """Returns a `Section` instance."""
    type = ""
    markdown = ""
    items = []
    if name in ["Parameters", "Attributes"]:
        items = [Item(n, t, m) for n, t, m in parse_parameters(doc, style)]
    elif name == "Raises":
        items = [Item(t, "", m) for t, m in parse_raises(doc, style)]
    elif name in ["Returns", "Yields"]:
        type, markdown = parse_returns(doc, style)
    else:
        markdown = doc
    return Section(name, type=type, markdown=markdown, items=items)


def parse_property(doc: Docstring, obj: Any):
    """Parses property's docstring to inspect type."""
    section = doc.sections[0]
    markdown = section.markdown
    line = markdown.split("\n")[0]
    if ":" in line:
        index = line.index(":")
        doc.type = line[:index].strip()
        section.markdown = markdown[index + 1 :].strip()
    if not doc.type and hasattr(obj, "fget"):
        doc.type = Signature(obj.fget).returns


def postprocess(doc: Docstring, obj: Any):
    if isinstance(obj, property):
        parse_property(doc, obj)
    if not callable(obj):
        return
    signature = Signature(obj)
    if signature.signature is None:
        return

    if doc["Parameters"] is not None:
        for item in doc["Parameters"]:
            if item.type == "" and item.name in signature:
                item.type = signature.parameters[item.name]
                if item.type.startswith("("):  # tuple
                    item.type = item.type[1:-1]
            if "{default}" in item.markdown and item.name in signature:
                default = signature.defaults[item.name]
                item.markdown = item.markdown.replace("{default}", default)

    if doc["Attributes"] is not None and signature.attributes:
        for item in doc["Attributes"]:
            if item.type == "" and item.name in signature.attributes:
                item.type = signature.attributes[item.name]
                if item.type.startswith("("):
                    item.type = item.type[1:-1]

    for name in ["Returns", "Yields"]:
        section = doc[name]
        if section is not None and section.type == "":
            section.type = getattr(signature, name.lower())

    if doc["Returns"] is None and doc["Yields"] is None:
        from mkapi.core.node import get_kind

        kind = get_kind(obj)
        if kind == "generator":
            doc.type = signature.yields
        else:
            doc.type = signature.returns


def get_docstring(obj: Any) -> Docstring:
    """Returns a `Docstring` instance."""
    doc = inspect.getdoc(obj)
    if not doc:
        return Docstring()
    sections = []
    for section in split_section(doc):
        sections.append(create_section(*section))
    docstring = Docstring(sections)
    postprocess(docstring, obj)
    return docstring



import inspect
from dataclasses import dataclass, field
from functools import lru_cache
from typing import Any, Dict, Mapping, Optional, Union


@dataclass
class Signature:
    obj: Any = field(default=None, repr=False)
    signature: Optional[inspect.Signature] = None
    parameters: Mapping[str, str] = field(default_factory=dict, init=False)
    defaults: Dict[str, Any] = field(default_factory=dict, init=False)
    attributes: Dict[str, str] = field(default_factory=dict, init=False)
    returns: str = ""
    yields: str = ""

    def __post_init__(self):
        if self.obj is None or not callable(self.obj):
            return
        try:
            self.signature = inspect.signature(self.obj)
        except (TypeError, ValueError):
            pass

        for name, parameter in self.signature.parameters.items():
            if name == "self":
                continue
            type = to_string(parameter.annotation)
            default = parameter.default
            if default == inspect.Parameter.empty:
                self.defaults[name] = default
            else:
                self.defaults[name] = f"{default!r}"
                if not type.endswith(", optional"):
                    type += ", optional"
            self.parameters[name] = type

        return_annotation = self.signature.return_annotation
        self.returns = to_string(return_annotation, "returns")
        self.yields = to_string(return_annotation, "yields")

        if hasattr(self.obj, "__dataclass_fields__"):
            values = self.obj.__dataclass_fields__.values()
            self.attributes = {f.name: to_string(f.type) for f in values}

    def __contains__(self, name):
        return name in self.parameters

    def __getitem__(self, name):
        return getattr(self, name.lower())

    def __str__(self):
        args = []
        for arg in self.parameters:
            if self.defaults[arg] != inspect.Parameter.empty:
                arg += "=" + self.defaults[arg]
            args.append(arg)
        return "(" + ", ".join(args) + ")"


def to_string(annotation, kind: str = "") -> str:
    if kind == "yields":
        if hasattr(annotation, "__args__") and annotation.__args__:
            return to_string(annotation.__args__[0])
        else:
            return ""

    if hasattr(annotation, "__forward_arg__"):
        return annotation.__forward_arg__
    if annotation == inspect.Parameter.empty or annotation is None:
        return ""
    if hasattr(annotation, "__name__"):
        name = annotation.__name__
        if not hasattr(annotation, "__module__"):
            return name
        module = annotation.__module__
        if module == "builtins":
            return name
        return f"[{name}]({module}.{name})"
    if not hasattr(annotation, "__origin__"):
        return str(annotation).replace("typing.", "")
    origin = annotation.__origin__
    if origin is Union:
        return union(annotation)
    if origin is tuple:
        args = [to_string(x) for x in annotation.__args__]
        if args:
            return "(" + ", ".join(args) + ")"
        else:
            return "tuple"
    if origin is dict:
        args = [to_string(x) for x in annotation.__args__]
        if args:
            return "dict(" + ": ".join(args) + ")"
        else:
            return "dict"
    if hasattr(annotation, "__args__") and len(annotation.__args__) <= 1:
        return a_of_b(annotation)
    return ""


def a_of_b(annotation) -> str:
    """Returns A of B style string.

    Examples:
        >>> from typing import List, Iterable, Iterator
        >>> a = List[str]
        >>> a_of_b(a)
        'list of str'
        >>> a = Iterable[int]
        >>> a_of_b(a)
        'iterable of int'
        >>> a = Iterator[float]
        >>> a_of_b(a)
        'iterator of float'
    """
    origin = annotation.__origin__
    if not hasattr(origin, "__name__"):
        return ""
    name = origin.__name__.lower()
    type = f"{name} of " + to_string(annotation.__args__[0])
    if type.endswith(" of T"):
        return name
    return type


def union(annotation) -> str:
    """Returns a string for union annotation.

    Examples:
        >>> from typing import List, Optional, Tuple, Union
        >>> a = Optional[List[str]]
        >>> union(a)
        'list of str, optional'
        >>> a = Union[str, int]
        >>> union(a)
        'str or int'
        >>> a = Union[str, int, float]
        >>> union(a)
        'str, int, or float'
        >>> a = Union[List[str], Tuple[int, int]]
        >>> union(a)
        'Union(list of str, (int, int))'
    """
    args = annotation.__args__
    if (
        len(args) == 2
        and hasattr(args[1], "__name__")
        and args[1].__name__ == "NoneType"
    ):
        return to_string(args[0]) + ", optional"
    else:
        args = [to_string(x) for x in args]
        if all(" " not in arg for arg in args):
            if len(args) == 2:
                return " or ".join(args)
            else:
                return ", ".join(args[:-1]) + ", or " + args[-1]
        else:
            return "Union(" + ", ".join(to_string(x) for x in args) + ")"


@lru_cache(maxsize=1000)
def get_signature(obj: Any) -> Signature:
    return Signature(obj)



from typing import Dict, Tuple

import mkapi
from mkapi.core.base import Item, Node, Section


def get_params(node: Node, name: str) -> Tuple[Dict[str, str], Dict[str, str]]:
    section = node.docstring[name]
    if section is None:
        docstring_params = {}
    else:
        docstring_params = {item.name: item.type for item in section}
    signature_params = node.signature[name]
    return docstring_params, signature_params


def is_complete(node: Node, name: str = "") -> bool:
    if not name:
        return all(is_complete(node, name) for name in ["Parameters", "Attributes"])

    docstring_params, signature_params = get_params(node, name)
    for param in signature_params:
        if param not in docstring_params:
            return False
    return True


def inherit_base(node: Node, base: Node, name: str = ""):
    if not name:
        for name in ["Parameters", "Attributes"]:
            inherit_base(node, base, name)
        return

    base_section = base.docstring[name]
    if base_section is None:
        return
    _, node_params = get_params(node, name)
    _, base_params = get_params(base, name)
    node_section = node.docstring[name]
    items = []
    for item in base_section:
        if node_section is None or item.name not in node_section:
            if (
                item.name in node_params
                and node_params[item.name] == base_params[item.name]
            ):
                items.append(item)
    if node_section is not None:
        for item in node_section:
            if item not in items:
                items.append(item)
    node.docstring[name] = Section(name, items=items)  # type:ignore


def inherit_signature(node: Node, name: str = ""):
    if not name:
        for name in ["Parameters", "Attributes"]:
            inherit_signature(node, name)
        return

    _, params = get_params(node, name)
    if not params:
        return

    node_section = node.docstring[name]
    if node_section is None:
        items = []
    else:
        items = node_section.items
    for item_name, type in params.items():
        if node_section is None or item_name not in node_section:
            item = Item(item_name, type, markdown="MISSING")
            items.append(item)
    node.docstring[name] = Section(name, items=items)  # type:ignore


def inherit(node: Node):
    if is_complete(node):
        return
    for cls in node.obj.mro()[:-1]:
        base = mkapi.get_node(cls)
        inherit_base(node, base)
        if is_complete(node):
            return
    inherit_signature(node)



import os
import re
from typing import List

from mkapi.core.base import Base, Node

LINK_PATTERN = re.compile(r"\[(.+?)\]\((.+?)\)")


def resolve_link(node: Node, abs_src_path: str, api_roots: List[str]):
    resolve_node_link(node, abs_src_path, api_roots)
    for base in node:
        resolve_type_link(base, abs_src_path, api_roots)


def resolve_node_link(node: Node, abs_src_path: str, api_roots: List[str]):
    node.prefix_url = resolve_href(node.prefix, abs_src_path, api_roots, is_html=True)
    if node.prefix_url:
        node.name_url = ".".join([node.prefix_url, node.name])
    resolve_type_link(node, abs_src_path, api_roots)
    for member in node.members:
        resolve_node_link(member, abs_src_path, api_roots)


def resolve_type_link(base: Base, abs_src_path: str, api_roots: List[str]):
    def replace(match):
        name = match.group(1)
        href = resolve_href(match.group(2), abs_src_path, api_roots, is_html=True)
        if href:
            title = href.split("#")[1]
            return f'<a href="{href}" title="{title}">{match.group(1)}</a>'
        else:
            return name

    base.type_html = re.sub(LINK_PATTERN, replace, base.type)


def resolve_markdown_link(
    markdown: str, abs_src_path: str, api_roots: List[str]
) -> str:
    def replace(match):
        name = match.group(1)
        if match.group(2).startswith("!"):  # Just for MkApi documentation.
            href = match.group(2)[1:]
            return f"[{name}]({href})"
        href = resolve_href(match.group(2), abs_src_path, api_roots, is_html=False)
        if href:
            return f"[{name}]({href})"
        else:
            return match.group()

    return re.sub(LINK_PATTERN, replace, markdown)


def resolve_href(
    name: str, abs_src_path: str, api_roots: List[str], is_html: bool
) -> str:
    if not name:
        return ""
    for root in api_roots:
        dirname, path = os.path.split(root)
        if name.startswith(path[:-3]):
            if is_html:
                relpath = os.path.relpath(root, abs_src_path)
            else:
                relpath = os.path.relpath(root, os.path.dirname(abs_src_path))
            if is_html:
                relpath = relpath[:-3]
            relpath = relpath.replace("\\", "/")
            return "#".join([relpath, name])
    return ""



import importlib
import inspect
from functools import lru_cache, partial
from typing import Any, List, Tuple

from mkapi.core.base import Node
from mkapi.core.docstring import get_docstring
from mkapi.core.signature import get_signature

ISFUNCTIONS = {}
for x in dir(inspect):
    if x.startswith("is"):
        name = x[2:]
        if name not in ["routine", "builtin", "code"]:
            ISFUNCTIONS[name] = getattr(inspect, x)


def get_kinds(obj) -> List[str]:
    kinds = []
    for kind, func in ISFUNCTIONS.items():
        if func(obj):
            kinds.append(kind)
    return kinds


def get_kind(obj) -> str:
    if hasattr(obj, "__dataclass_fields__"):
        return "dataclass"
    if isinstance(obj, property):
        if obj.fset:
            return "readwrite_property"
        else:
            return "readonly_property"
    kinds = get_kinds(obj)
    if not kinds:
        return ""
    if "generatorfunction" in kinds:
        return "generator"
    if "function" in kinds:
        try:
            parameters = inspect.signature(obj).parameters
        except (ValueError, TypeError):
            return ""
        if parameters:
            arg = list(parameters)[0]
            if arg == "self":
                return "method"
    kind = kinds[-1]
    if kind == "module":
        try:
            sourcefile = inspect.getsourcefile(obj)
        except TypeError:
            return ""
        if sourcefile and sourcefile.endswith("__init__.py"):
            kind = "package"
    return kind


def get_sourcefile_and_lineno(obj) -> Tuple[str, int]:
    if isinstance(obj, property):
        obj = obj.fget
    try:
        sourcefile = inspect.getsourcefile(obj) or ""
        lineno = inspect.getsourcelines(obj)[1]
    except (TypeError, OSError):
        return "", -1
    return sourcefile, lineno


def filter(obj, qualname, sourcefile="") -> bool:
    if isinstance(obj, property):
        return True
    kind = get_kind(obj)
    if kind == "":
        return False
    if kind == 'method':
        return True

    sourcefile_, _ = get_sourcefile_and_lineno(obj)
    if sourcefile_ == "" or (sourcefile and sourcefile != sourcefile_):
        return False
    if hasattr(obj, "__qualname__"):
        if not qualname:
            return True
        if obj.__qualname__.startswith(qualname):
            return True
    return False


def get_members(obj, kind, sourcefile, prefix, depth, max_depth=-1) -> List[Node]:
    if max_depth != -1 and depth > max_depth:
        return []
    if isinstance(obj, property):
        return []

    qualname = getattr(obj, "__qualname__", "")
    if kind in ["package", "module"]:
        func = partial(filter, qualname=qualname, sourcefile=sourcefile)
    else:
        func = partial(filter, qualname=qualname)

    members = []
    for name, obj in inspect.getmembers(obj, func):
        if name.startswith("_") and name != "__init__":
            continue
        member = walk(obj, prefix, name, depth=depth, max_depth=max_depth)
        # Below is needed for max_depth.
        if member.kind in ["class", "dataclass"] and not member.docstring:
            docstring = get_docstring(obj.__init__)
            if docstring:
                markdown = docstring.sections[0].markdown
                if not markdown.startswith("Initialize self"):
                    member.docstring = docstring
        if member.docstring:
            members.append(member)
    return sorted(members, key=lambda x: (x.sourcefile, x.lineno))


def walk(obj, prefix, name, depth=0, max_depth=-1) -> Node:
    member_prefix = name
    if prefix:
        member_prefix = ".".join([prefix, member_prefix])
    kind = get_kind(obj)
    sourcefile, lineno = get_sourcefile_and_lineno(obj)
    signature = get_signature(obj)
    docstring = get_docstring(obj)
    members = get_members(obj, kind, sourcefile, member_prefix, depth + 1, max_depth)

    if kind in ["class", "dataclass"] and not docstring:
        for member in members:
            if member.name == "__init__" and member.docstring:
                markdown = member.docstring.sections[0].markdown
                if not markdown.startswith("Initialize self"):
                    docstring = member.docstring
    members = [member for member in members if member.name != "__init__"]

    node = Node(
        obj=obj,
        prefix=prefix,
        name=name,
        depth=depth,
        kind=kind,
        sourcefile=sourcefile,
        lineno=lineno,
        signature=signature,
        docstring=docstring,
        members=members,
    )
    if docstring and docstring.type:
        node.type = docstring.type
    return node


def get_object(name: str) -> Any:
    """Reutrns an object specified `name`.

    Examples:
        >>> import inspect
        >>> obj = get_object('mkapi.core')
        >>> inspect.ismodule(obj)
        True
        >>> obj = get_object('mkapi.core.base')
        >>> inspect.ismodule(obj)
        True
        >>> obj = get_object('mkapi.core.base.Node')
        >>> inspect.isclass(obj)
        True
        >>> obj = get_object('mkapi.core.base.Node.get_markdown')
        >>> inspect.isfunction(obj)
        True
    """
    names = name.split(".")
    for k in range(len(names), 0, -1):
        module_name = ".".join(names[:k])
        try:
            obj = importlib.import_module(module_name)
        except ModuleNotFoundError:
            continue
        for attr in names[k:]:
            obj = getattr(obj, attr)
        return obj
    raise ValueError(f"Could not find object: {name}")


def split_prefix_and_basename(obj) -> Tuple[str, str]:
    """Split an object name into prefix and basename.

    Examples:
        >>> import inspect
        >>> obj = get_object('mkapi.core')
        >>> split_prefix_and_basename(obj)
        ('', 'mkapi.core')
        >>> obj = get_object('mkapi.core.base')
        >>> split_prefix_and_basename(obj)
        ('', 'mkapi.core.base')
        >>> obj = get_object('mkapi.core.base.Node')
        >>> split_prefix_and_basename(obj)
        ('mkapi.core.base', 'Node')
        >>> obj = get_object('mkapi.core.base.Node.get_markdown')
        >>> split_prefix_and_basename(obj)
        ('mkapi.core.base.Node', 'get_markdown')
    """
    if inspect.ismodule(obj):
        return "", obj.__name__
    else:
        module = obj.__module__
        qualname = obj.__qualname__
        if "." not in qualname:
            prefix, basename = module, qualname
        else:
            prefix, _, basename = qualname.rpartition(".")
            prefix = ".".join([module, prefix])
        if prefix == "__main__":
            prefix = ""
        return prefix, basename


@lru_cache(maxsize=1000)
def get_node(name: Any, max_depth: int = -1, headless: bool = False) -> Node:
    if isinstance(name, str):
        obj = get_object(name)
    else:
        obj = name
    prefix, basename = split_prefix_and_basename(obj)
    node = walk(obj, prefix, basename, max_depth=max_depth)
    node.headless = headless
    return node



import re
from dataclasses import dataclass, field
from typing import Iterator, List

from mkapi.core.inherit import inherit
from mkapi.core.linker import resolve_link, resolve_markdown_link
from mkapi.core.node import Node, get_node
from mkapi.core.renderer import renderer

MKAPI_PATTERN = re.compile(r"^!\[mkapi\]\((.+?)\)$", re.MULTILINE)

HTML_PATTERN = re.compile(
    r"<!-- mkapi:(\d+):begin -->(.*?)<!-- mkapi:end -->", re.MULTILINE | re.DOTALL
)


@dataclass
class Page:
    source: str
    abs_src_path: str
    api_roots: List[str] = field(default_factory=list)
    nodes: List[Node] = field(default_factory=list, init=False)

    def __post_init__(self):
        markdown = "\n\n".join(self.split(self.source))
        self.markdown = resolve_markdown_link(
            markdown, self.abs_src_path, self.api_roots
        )

    def split(self, source) -> Iterator[str]:
        cursor = 0
        for index, match in enumerate(MKAPI_PATTERN.finditer(source)):
            start, end = match.start(), match.end()
            if cursor < start:
                markdown = source[cursor:start].strip()
                if markdown:
                    yield markdown
            name = match.group(1)
            if ":" in name:
                name, max_depth_str = name.split(":")
                max_depth = int(max_depth_str)
            else:
                max_depth = -1
            if name[0] == "#":
                headless = True
                name = name[1:]
            else:
                headless = False
            node = get_node(name, max_depth, headless)
            if node.kind in ["class", "dataclass"]:
                inherit(node)
            resolve_link(node, self.abs_src_path, self.api_roots)
            self.nodes.append(node)
            markdown = node.get_markdown()
            yield f"<!-- mkapi:{index}:begin -->\n\n{markdown}\n\n<!-- mkapi:end -->"
            cursor = end
        if cursor < len(source):
            markdown = source[cursor:].strip()
            if markdown:
                yield markdown

    def content(self, html):
        def replace(match):
            node = self.nodes[int(match.group(1))]
            node.set_html(match.group(2))
            return renderer.render(node)

        return re.sub(HTML_PATTERN, replace, html)
